import numpy as nm

from sfepy.base.base import assert_, Struct
from sfepy.terms.terms import Term, terms

class CouplingVectorScalarTL(Struct):
    
    def set_data_shape(self, apv, aps):
        """Set element data shape and checks dimensions of approximations."""
        if self.dof_conn_type == 'surface':
            self.data_shape_v = apv.get_s_data_shape(self.integral,
                                                     self.region.name)
            self.data_shape_s = aps.get_s_data_shape(self.integral,
                                                     self.region.name)
        else:
            self.data_shape_v = apv.get_v_data_shape(self.integral)
            self.data_shape_s = aps.get_v_data_shape(self.integral)

        assert_(aps.dim == (1,))
        assert_(apv.dim == (self.data_shape_v[2],))

    def get_shape_div(self, diff_var, chunk_size):
        n_el, n_qp, dim, n_eps = self.data_shape_s

        if diff_var is None:
            return (chunk_size, 1, n_eps, 1), 0
        elif diff_var == self.get_arg_name( 'state_p' ):
            return (chunk_size, 1, n_eps, n_eps), 2
        elif diff_var == self.get_arg_name( 'state' ):
            n_epv = self.data_shape_v[3]
            return (chunk_size, 1, n_eps, dim * n_epv), 1
        else:
            raise StopIteration

    def get_shape_grad(self, diff_var, chunk_size):
        n_el, n_qp, dim, n_epv = self.data_shape_v

        if diff_var is None:
            return (chunk_size, 1, dim * n_epv, 1), 0
        elif diff_var == self.get_arg_name( 'state' ):
            return (chunk_size, 1, dim * n_epv, dim * n_epv), 1
        elif diff_var == self.get_arg_name( 'state_p' ):
            n_eps = self.data_shape_s[3]
            return (chunk_size, 1, dim * n_epv, n_eps), 2
        else:
            raise StopIteration

class HyperElasticBase( Term ):
    """Base class for all hyperelastic terms in TL/UL formulation.

    **Note** This is not a proper Term!

    `HyperElasticBase.__call__()` computes element contributions given either
    stress (-> rezidual) or tangent modulus (-> tangent sitffnes matrix),
    i.e. constitutive relation type (CRT) related data. The CRT data are
    computed in subclasses implementing particular CRT (e.g. neo-Hookean
    material), in self.compute_crt_data().
    Mode: 0 - total formulation, 1 - updated formulation
    """
    def __init__(self, *args, **kwargs):
        Term.__init__(self, *args, **kwargs)

        self.mode_ul = {'tl' : 0, 'ul' : 1}[self.mode]
        self.function = {
            'finite_strain' : { 0: 'finite_strain_tl',
                                1: 'finite_strain_ul'},
            'element_contribution' : terms.dw_he_rtm,
        }

        igs = self.region.igs
        dummy = nm.array([0], ndmin=4)
        self.crt_data = Struct(stress={}.fromkeys(igs, None),
                               tan_mod=dummy)

    def __call__( self, diff_var = None, chunk_size = None, **kwargs ):
        term_mode, = self.get_kwargs( ['term_mode'], **kwargs )
        virtual, state = self.get_args( ['virtual', 'state'], **kwargs )
        ap, vg = self.get_approximation(virtual)

        self.set_data_shape(ap)
        shape, mode = self.get_shape(diff_var, chunk_size)

        cache = self.get_cache( self.function['finite_strain'][self.mode_ul], 0 )
        family_data = cache(self.family_data_names, self, 0, state=state)
##         print family_data

        if term_mode is None:
            ig = self.char_fun.ig

            out = self.compute_crt_data( family_data, mode, **kwargs )

            if mode == 0:
                self.crt_data.stress[ig] = stress = out

            else:
                self.crt_data.tan_mod = out

                stress = self.crt_data.stress[ig]
                if stress is None:
                    stress = self.compute_crt_data(family_data, 0, **kwargs)

            fun = self.function['element_contribution']

            mtxF, detF = cache(['F', 'detF'], self, 0, state=state)

            for out, chunk in self.char_fun( chunk_size, shape ):
                status = fun( out, stress, self.crt_data.tan_mod,
                              mtxF, detF, vg, chunk, mode, self.mode_ul )
                yield out, chunk, status

        elif term_mode == 'd_eval':
            raise NotImplementedError

        elif term_mode in ['strain', 'stress']:

            if term_mode == 'strain':
                out_qp = cache('E', self, 0, state=state)

            elif term_mode == 'stress':
                out_qp = self.compute_crt_data( family_data, 0, **kwargs )
                
            shape = (chunk_size, 1) + out_qp.shape[2:]
            for out, chunk in self.char_fun( chunk_size, shape ):
                status = vg.integrate_chunk( out, out_qp[chunk], chunk )
                out1 = out / vg.variable( 2 )[chunk]

            yield out1, chunk, status

class DeformationGradientTerm(Term):
    r"""
    :Description:
    Deformation gradient :math:`F` in quadrature points for
    `term_mode='def_grad'` (default) or the jacobian :math:`J` if
    `term_mode='jacobian'`.

    :Definition:
    .. math::
        \ull{F} = \pdiff{\ul{x}}{\ul{X}}|_{qp}
        = \ull{I} + \pdiff{\ul{u}}{\ul{X}}|_{qp} \;, \\
        \ul{x} = \ul{X} + \ul{u} \;, J = \det{(\ull{F})}

    :Arguments:
        state : :math:`\ul{u}`
    """
    name = 'dq_def_grad'
    arg_types = ('state',)

    function = staticmethod(terms.dq_def_grad)

    def __call__(self, diff_var=None, chunk_size=None, **kwargs):
        state, = self.get_args(**kwargs)
        term_mode = kwargs.get('term_mode', 'dq_def_grad')

        ap, vg = self.get_approximation(state)
        n_el, n_qp, dim, n_ep = ap.get_v_data_shape(self.integral)

        if diff_var is None:
            if term_mode == 'def_grad':
                shape = (chunk_size, n_qp, dim, dim)
                mode = 0

            elif term_mode == 'jacobian':
                shape = (chunk_size, n_qp, 1, 1)
                mode = 1

        else:
            raise StopIteration

        vec = state()
        for out, chunk in self.char_fun(chunk_size, shape):
            status = self.function(out, vec, vg, ap.econn, chunk, mode)
            yield out, chunk, status
